#include "Constants.h"
#include "SCDVK.h"
#include <Wire.h>
#include <SPI.h>
#include <SD.h>

#define debug

#ifdef debug
#define d_println(x) SerialUSB.println(x); d_echo(); delay(3)
#define d_print(x) SerialUSB.print(x)
#else
#define d_println(x) delay(3)
#define d_print(x)
#endif

#define SPI_TX_BUFFER 1024
#define SPI_RX_BUFFER 256
#define SPI_SPEED 1000000
#define SPI_TIMEOUT 3000 //ms

//operation mode: 0=normal operation   4=webserver mode (webserver by default enabled in rtx)
int operation_mode = 0;
//int operation_mode = 4;


/*****SPI vars*****/

//modified by irq
volatile bool rtx_signal_off = false;
volatile bool rtx_signal_on = false;

//buffer
byte SPI_tx[SPI_TX_BUFFER];
byte SPI_rx[SPI_RX_BUFFER];
byte SPI_rx_count = 0;


/**********************others**********************/
bool SD_INIT = false;            //sd initialized?
bool softap_enabled = false;     //softap enabled (rtx)?
bool webserver_enabled = false;  //webserver enabled (rtx)?
bool timeout_error = false;      //error?
unsigned long timeisr = 0;       //timer for some irq
unsigned long time_timeout = 0;  //timer for spi timeouts



/**********************IRQ**********************/
void ISR1()
{
  if ((millis() - timeisr) > 1000)
  {
    timeisr = millis();
    if (operation_mode == 1) operation_mode = 2;
    else if (operation_mode == 0) operation_mode = 1;
  }
}

void ISR2()
{
  if ((millis() - timeisr) > 1000)
  {
    timeisr = millis();
    if (operation_mode == 2) operation_mode = 3;
  }
}

void ISR_RTX_SPI()
{
  bool s = digitalRead(RTX_IRQ_SPI);
  if (s) rtx_signal_off = true;
  else rtx_signal_on = true;
}
/****************************************************/



/***************** HELPER FUNCTIONS *****************/

/*!
 * @brief Convert a long var to an array of bytes
 * @param dst: Ptr to array of bytes
 * @param dec_num: Number to convert
 * @return number of bytes required
 */
byte dec_to_hex(byte * dst, long dec_num)
{
  byte l = 0;
  byte res[5]; //5 bytes for store a long var
  memset(&res[0], 0, 5);

  while (dec_num > 0)  {
    res[l] = dec_num % 256;
    dec_num >>= 8;
    l++;
  }
  memcpy(dst, &res[0], sizeof(res));
  return l;
}

/*!
 * @brief Read N bytes from SPI
 * @param l: Number bytes to read
 */
void connread_rtx_spi(int l)
{
  SPI_rx_count = 0;
  digitalWrite(RTX_CS, HIGH);
  SPI.beginTransaction(SPISettings(SPI_SPEED, LSBFIRST, SPI_MODE0));
  for (int i = 0; i < l; i++)
  {
    SPI_rx[SPI_rx_count] = SPI.transfer(0x00);
    SPI_rx_count ++;
  }
  SPI.endTransaction();
  digitalWrite(RTX_CS, LOW);
}

/*!
 * @brief Send N bytes to SPI
 * @param data: Ptr to array
 * @param len: Number bytes to send
 */
void connsend_rtx_spi(byte * data, int len)
{
  digitalWrite(RTX_CS, HIGH);
  SPI.beginTransaction(SPISettings(SPI_SPEED, LSBFIRST, SPI_MODE0));
  byte s;
  for (int i = 0; i < len; i++)
  {
    memcpy (&s, data + i, 1);
    SPI.transfer((s));
  }
  SPI.endTransaction();
  digitalWrite(RTX_CS, LOW);
  delay (10);
}


/*!
 * @brief Initialize variables, pin modes, etc
 */
void SCDVK::begin() {
  pinMode(RESET, OUTPUT);
  digitalWrite(RESET, LOW); //RESET ON
  delay(100);
  digitalWrite(RESET, HIGH); //RESET OFF
  pinMode(MUX, OUTPUT);
  digitalWrite(MUX, HIGH); //FIRMWARE MODE
  pinMode(PROG_MODE, OUTPUT);
  digitalWrite(PROG_MODE, LOW); //PROG_MODE OFF
  pinMode(GREEN, OUTPUT);
  digitalWrite(GREEN, HIGH); //GREEN ON
  pinMode(BLUE, OUTPUT);
  digitalWrite(BLUE, LOW); //BLUE OFF
  pinMode(MMC_CS, OUTPUT);
  pinMode(52, OUTPUT);
  digitalWrite(MMC_CS, HIGH); //SPI MMC NO SELECT
  pinMode(RTX_CS, OUTPUT);
  digitalWrite(RTX_CS, LOW); //SPI RTX NO SELECT
  pinMode(POWER, OUTPUT);
  digitalWrite(POWER, HIGH); //ENABLE POWER RTX
  pinMode(SELECT_MODE_1, INPUT);
  pinMode(SELECT_MODE_2, INPUT);
  pinMode(RTX_IRQ_SPI, INPUT);
  Serial.begin(115200);
  Serial1.begin(115200);
  SerialUSB.begin(9600);
  Wire.begin();
  attachInterrupt(SELECT_MODE_1, ISR1, FALLING) ;
  attachInterrupt(SELECT_MODE_2, ISR2, FALLING) ;
  SD_INIT = (SD.begin(MMC_CS));  //Start SD (include: SPI.BEGIN)
}

/*!
 * @brief Send API(SPI) to start WebServer and prepare for it
 */
void SCDVK::webserver_mode()
{
  digitalWrite(GREEN, HIGH); //GREEN LED OFF
  digitalWrite(BLUE, HIGH); //BLUE LED ON
  while (operation_mode == 2)
  {
    if (!webserver_enabled)
    {
      attachInterrupt(RTX_IRQ_SPI, ISR_RTX_SPI, CHANGE);
      d_print("Starting webserver...");
      SPI_tx[0] = 0x11;
      connsend_rtx_spi(&SPI_tx[0], 1);

      while (!rtx_signal_off); //starting
      while (!rtx_signal_on) d_echo(); //when start -> finished

      webserver_enabled = true;
      d_println("OK");
      d_echo();
      rtx_signal_off = false;
      rtx_signal_on = false;
    }
    echo();
    if (rtx_signal_off == true) rtxserver();
  }
}

/*!
 * @brief Send API(SPI) and exit webserver mode
 */
void SCDVK::webserver_exit()
{
  if (webserver_enabled)
  {
    d_println("Stopping webserver & SOFT AP... (0x12)");
    SPI_tx[0] = 0x12;
    connsend_rtx_spi(&SPI_tx[0], 1);
    webserver_enabled = false;
    softap_enabled = false;
    detachInterrupt(RTX_IRQ_SPI);
    operation_mode = 0;
  }
}

/*!
 * @brief Normal execution (without softap)
 */
void SCDVK::normal_mode()
{
  digitalWrite(GREEN, LOW); //GREEN LED OFF
  digitalWrite(BLUE, LOW); //GREEN LED OFF
  while (operation_mode == 0)
  {
    main();
    echo();
  }
}

/*!
 * @brief Send API(SPI) to start SoftAP
 */
void SCDVK::softap_mode()
{
  digitalWrite(GREEN, HIGH); //GREEN LED ON
  digitalWrite(BLUE, LOW); //GREEN LED OFF
  //  static byte test = 0;
  while (operation_mode == 1)
  {
    if (!softap_enabled)
    {
      attachInterrupt(RTX_IRQ_SPI, ISR_RTX_SPI, CHANGE);
      d_print("Starting SOFT AP... ");
      SPI_tx[0] = 0x10;
      connsend_rtx_spi(&SPI_tx[0], 1);

      while (!rtx_signal_off); //starting
      //while (digitalRead(RTX_IRQ_SPI) == HIGH) d_echo(); //when start -> finished
      while (!rtx_signal_on); //when start -> finished

      softap_enabled = true;
      detachInterrupt(RTX_IRQ_SPI);
      rtx_signal_off = false;
      rtx_signal_on = false;
      d_println("OK");
    }
    echo();
  }
}

/*!
 * @brief Send data btwn serials(1&2)
 */
void SCDVK::echo()
{
  if (Serial.available())
    Serial1.write(Serial.read());
  if (Serial1.available())
    Serial.write(Serial1.read());
}

/*!
 * @brief Force serverweb state (without start it)
 */
void SCDVK::test_serverweb()
{
  digitalWrite(BLUE, HIGH);
  attachInterrupt(RTX_IRQ_SPI, ISR_RTX_SPI, CHANGE);

  while (1)
  {
    echo();
    if (rtx_signal_off == true) rtxserver();
  }
}

/*!
 * @brief executed program (loop)
 */
void SCDVK::execute()
{
  switch (operation_mode) {
    case 0:
      normal_mode();
      break;
    case 1:
      softap_mode();
      break;
    case 2:
      webserver_mode();
      break;
    case 3:
      webserver_exit();
      break;
    case 4:
      test_serverweb();
      break;
  }
}

/*!
 * @brief main for default mode
 */
void SCDVK::main()
{
  if (webserver_enabled) {
    d_println("Stopping webserver");
    webserver_enabled = false;
  }

  if (softap_enabled) {
    d_println("Stopping softap");
    softap_enabled = false;
  }

}

/*!
 * @brief Send ALL data btwn serials(1&2)
 * @param cmd: Ptr to cmd to send
 * @param data: Ptr to save result
 */
void SCDVK::exec_cmd(char *cmd, char *data)
{
  //  SCDVK DVK;
  byte l = 0;

  if (strcmp(cmd, "getHumidity") == 0) {
    //    sprintf(data + 1, "%f", DVK.getHumidity());
    sprintf(data + 1, "%f", getHumidity());
    l = 5;
    //    d_print("Q hum -> "); d_println(DVK.getHumidity());
    d_print("Q hum -> "); d_println(getHumidity());
  }
  else if (strcmp(cmd, "getTemperature") == 0) {
    //    sprintf(data + 1, "%f", DVK.getTemperature());
    sprintf(data + 1, "%f", getTemperature());
    l = 5;
    //    d_print("Q temp -> "); d_println(DVK.getTemperature());
    d_print("Q temp -> "); d_println(getTemperature());
  }
  //...
  else sprintf(data + 1, "Unknown command");

  SPI_tx[0] = (l > 0) ? l : strlen(data + 1);
}

/*!
 * @brief Process API rcv by RTX (webserver mode)
 */
void SCDVK::rtxserver()
{
  connread_rtx_spi(1);
  rtx_signal_off = false;

  d_print("CMD << RTX -> "); d_println(SPI_rx[0]);

  switch (SPI_rx[0]) {

    case 0xA0:
      {
        d_println("cmd request");

        char spi_cmd[256];

        connread_rtx_spi(1);
        byte size_cmd = SPI_rx[0];  //first byte = length path

        connread_rtx_spi(size_cmd);
        memcpy (&spi_cmd[0], &SPI_rx[0], size_cmd);
        spi_cmd[size_cmd] = 0x00;

        d_print("<<"); d_println(spi_cmd);

        exec_cmd(spi_cmd, (char*)&SPI_tx[0]);

        time_timeout = millis();
        //wait until rtx turn on led or timeout
        while (!rtx_signal_on) {
          if (millis() > (time_timeout + SPI_TIMEOUT)) {
            d_println("CMD_TIMEOUT");
            rtx_signal_off = false;
            rtx_signal_on = false;
            timeout_error = true;
            break;
          }
        }

        if (timeout_error) {
          timeout_error = false;  //timeout -> exit
          break;
        }

        rtx_signal_on = false;

        connsend_rtx_spi(&SPI_tx[0], SPI_tx[0] + 1);
        d_println("answer sended");
      }
      break;

    case 0xA1:
      {
        char spi_filename[256];
        d_println("file request");

        connread_rtx_spi(1);
        byte size_path = SPI_rx[0];  //first byte = length path

        connread_rtx_spi(size_path);
        memcpy (&spi_filename[0], &SPI_rx[0], size_path);
        spi_filename[size_path] = 0x00;
        while (!rtx_signal_on);
        rtx_signal_on = false;

        d_print("<<"); d_println(spi_filename);

        //sd available?
        if (!SD_INIT) SD_INIT = (SD.begin(MMC_CS));
        if (!SD_INIT) {
                    d_println("Err SD. -> File not exist");
                    connsend_rtx_spi((byte*)0x00, 1);
                    return ;
        }

        if (SD.exists(spi_filename)) {
          File f = SD.open(spi_filename);
          long r = f.size();

          byte bytes[5];
          byte nbytes = dec_to_hex(bytes, r);
          byte tosend[nbytes + 1];

          tosend[0] = nbytes; d_print("trying to read... ");
          memcpy( &tosend[1], &bytes[0], nbytes );

          connsend_rtx_spi(&tosend[0], sizeof(tosend)); //send length (SPI)
          d_print(">>size:"); d_println(r);

          while (!rtx_signal_off); //wait until rtx rcv
          rtx_signal_off = false;

          connread_rtx_spi(2);  //get buffer (always 2b)


          unsigned short spi_buffer = 0;

          spi_buffer += SPI_rx[0];
          spi_buffer += SPI_rx[1] * 256;

          //choose low buffer (SPI_TX_BUFFER) vs (spi_buffer)
          spi_buffer = min(spi_buffer, SPI_TX_BUFFER);
          d_print("Negotiated buffer: "); d_println(spi_buffer);
          byte bufferdata[spi_buffer];

          while (!rtx_signal_on);
          rtx_signal_on = false;


          dec_to_hex(&bytes[0], spi_buffer);
          connsend_rtx_spi(&bytes[0], 2);

          d_println("buffer sended...");

          while (f.available())
          {
            d_print("sending chunk file...");

            int i;
            for (i = 0; (i < spi_buffer) && f.available() ; i++) {
              bufferdata[i] = f.read();
            }

            while (!rtx_signal_off);
            rtx_signal_off = false;
            connsend_rtx_spi(&bufferdata[0], i);
            d_println("chunk sended. wait rtx confirmation...");
            time_timeout = millis();
            while (!((f.available() == 0) or rtx_signal_off)) { // while EOF or IRQ(RTX)
              if (millis() > (time_timeout + SPI_TIMEOUT)) {
                d_println("CMD_TIMEOUT");
                rtx_signal_off = false;
                rtx_signal_on = false;
                timeout_error = true;
                break;
              }
            }
            if (timeout_error) {
              timeout_error = false;  //timeout -> exit
              break;
            }
          }
          rtx_signal_off = false;
          rtx_signal_on = false;


          d_println("file sended, closing...");
          f.close();
        }
        else {
          d_println(">>not exist");
          connsend_rtx_spi((byte*)0x00, 1);
        }
      }
      break;

    default:
      d_println("CMD Unknown");
      rtx_signal_on = false;
  }
}

/*!
 * @brief Used for read data from SHT21
 * @param type: Type data called
 * @return number returned by SHT21
 */
uint16_t SCDVK::readSHT21(uint8_t type) {
  uint16_t DATA = 0;
  Wire.beginTransmission(SHT21_ADDRESS);
  Wire.write(type);
  Wire.endTransmission();
  Wire.requestFrom(SHT21_ADDRESS, 2);
  unsigned long time = millis();
  while (!Wire.available()) if ((millis() - time) > 500) return 0x00;
  DATA = Wire.read() << 8;
  while (!Wire.available());
  DATA = (DATA | Wire.read());
  DATA &= ~0x0003;
  return DATA;
}

/*!
 * @brief Used for get temperature
 * @return temperature in Cº
 */
float SCDVK::getTemperature()
{
  return  (-46.85 + ((175.72 * (float)readSHT21(0xE3)) / 65536.0) );
}

/*!
 * @brief Used for get humidity
 * @return humidity in %
 */
float SCDVK::getHumidity()
{
  return  (-6.0 + ((125.0  * (float)readSHT21(0xE5)) / 65536.0));
}

/*!
 * @brief Send ALL data btwn serials(1&2)
 */
void d_echo()
{
  while (Serial1.available() != 0 || Serial.available() != 0 )
  {
    if (Serial1.available()) Serial.write(Serial1.read());
    if (Serial.available()) Serial1.write(Serial.read());
  }
}
