/****************************************************************************
*  Program/file: DrvSpiSlave.c
*   
*  Copyright (C) by RTX A/S, Denmark.
*  These computer program listings and specifications, are the property of 
*  RTX A/S, Denmark and shall not be reproduced or copied or used in 
*  whole or in part without written permission from RTX A/S, Denmark.
*
*  Programmer: LKA
*
*  MODULE:
*  CONTROLLING DOCUMENT:
*  SYSTEM DEPENDENCIES:
*   
*  DESCRIPTION: Simple Slave Mode SPI driver. 
*               - Implemented for 8 bit data frames
*               - Support for 4 wire interface (CS active low)
*               - Implemented as SPI Mode 0 (CLKPOL = 0, CLKPHA = 0)
*               - Using USART2 of the MCU
*
****************************************************************************/

/****************************************************************************
*                                  PVCS info                                  
*****************************************************************************

$Author:   lka  $
$Date:   08 Mar 2013 10:31:02  $
$Revision:   1.0  $
$Modtime:   06 Mar 2013 23:16:58  $
$Archive:   J:/sw/Projects/Amelie/Components/Drivers/vcs/DrvSpiSlave.c_v  $

*/


/****************************************************************************
*                               Include files                                 
****************************************************************************/
//Type definitions
#include <Core/RtxCore.h> // Mandatory, Must be the first include

//Framework/Kernel
#include <Ros/RosCfg.h>   // The module sends mails or uses timers.
#include <PortDef.h>      // The module accesses the registers/HW/Interrupt 
                          // handlers. SOC is included via PortDef.h

//Interfaces
#include <Api/Api.h>
#include <Cola/cola.h>
#include <Protothreads/Protothreads.h>
#include <efm32.h>
#include <em_usart.h>
#include <em_cmu.h>

//Configurations

//Private 
#include <Drivers/DrvSpiSlave.h>
//#include <RtxEai/RtxEai.h>
/****************************************************************************
*                              Macro definitions                              
****************************************************************************/
  #define USART USART2
  #define USART_ROUTE_LOCATION USART_ROUTE_LOCATION_LOC0
  #define USART_RX_IRQN USART2_RX_IRQn
  #define USART_TX_IRQN USART2_TX_IRQn
  #ifdef _EFM32_GECKO_FAMILY
    #define USART_RX_INTERRUPT 30
    #define USART_TX_INTERRUPT 31
  #endif
  #ifdef _EFM32_GIANT_FAMILY
    #define USART_RX_INTERRUPT 34
    #define USART_TX_INTERRUPT 35
  #endif

#ifndef USART_SCU_DEVICE
  #define USART_SCU_DEVICE SCU_DEVICE2
#endif

#ifndef SPI_SLAVE_RX_BUFFER_SIZE
  #define SPI_SLAVE_RX_BUFFER_SIZE 512
#endif


/****************************************************************************
*                     Enumerations/Type definitions/Structs                   
****************************************************************************/
typedef struct
{
  rsbool Enabled;
  rsbool RxPending;
  rsuint16 RxWriteIdx;
  rsuint16 RxReadIdx;
  rsbool RxBofMailSent;
  rsuint8 *TxDataPtr;
  rsuint16 TxDataLength;
  rsuint16 TxCount;
} DrvSpiSlaveDataType;



/****************************************************************************
*                            Global variables/const                           
****************************************************************************/


/****************************************************************************
*                            Local variables/const                            
****************************************************************************/
// Static UART data
static DrvSpiSlaveDataType DrvSpiSlaveData;

// RX buffer
static rsuint8 SpiSlaveRxbuffer[SPI_SLAVE_RX_BUFFER_SIZE];

static rsuint16 trigger_buffer_rx;

/****************************************************************************
*                          Local Function prototypes                          
****************************************************************************/


/****************************************************************************
*                                Implementation                               
****************************************************************************/

// Calculates and returns the number of bytes stored in the RX buffer
rsuint16 DrvSpiSlaveRxGetSize(void)
{
  if (DrvSpiSlaveData.RxReadIdx > DrvSpiSlaveData.RxWriteIdx)
  {
    return SPI_SLAVE_RX_BUFFER_SIZE + DrvSpiSlaveData.RxWriteIdx - DrvSpiSlaveData.RxReadIdx;
  }
  else
  {
    return DrvSpiSlaveData.RxWriteIdx - DrvSpiSlaveData.RxReadIdx;
  }
}

// SPI MOSI interrupt handler 
void SpiSlaveRxIntHandler(rsuint8 vector)
{  
  rsuint32 usartif = USART_IntGet(USART);
  if (usartif & USART_IF_RXDATAV)
  {   
    // Read one byte from USART (this automatically clears the RXDATAV interrupt flag
    char c = USART_Rx(USART);
    
    // Calc new write idx
    rsuint16 idx = DrvSpiSlaveData.RxWriteIdx + 1;
    
    if (idx == SPI_SLAVE_RX_BUFFER_SIZE)
    {
      idx = 0;
    }

    if (idx == DrvSpiSlaveData.RxReadIdx)
    {
      // Buffer is full 
      
      // I not already done, send RX buffer overflow notification to application
      if (!DrvSpiSlaveData.RxBofMailSent)
      {
        DrvSpiSlaveData.RxBofMailSent = TRUE;
        RosMailSendP0(COLA_TASK, COLA_TASK, SPI_SLAVE_RX_BOF);
      }
     
      return; // to ignore this RX byte
    }
    
    // Put received byte in buffer
    SpiSlaveRxbuffer[DrvSpiSlaveData.RxWriteIdx] = c;
    DrvSpiSlaveData.RxWriteIdx = idx;
    
    // Notify the app if it hasn't already been notified
    if (DrvSpiSlaveData.RxPending == FALSE)
    {
      DrvSpiSlaveData.RxPending = TRUE;
      RosMailSendP0(COLA_TASK, COLA_TASK, SPI_SLAVE_RX_DATA);
    }  
    if (trigger_buffer_rx > 0 && DrvSpiSlaveData.RxWriteIdx == trigger_buffer_rx) RosMailSendP0(COLA_TASK, COLA_TASK, SPI_SLAVE_RX_TRIGGER);
  }
}

// SPI MISO interrupt handler
void SpiSlaveTxIntHandler(rsuint8 vector)
{
  rsuint32 usartif = USART_IntGet(USART);
  // Check TX Buffer Low interrupt flag
  if (usartif & USART_IF_TXBL)
  {
    
    // More to send?
    if (DrvSpiSlaveData.TxCount < DrvSpiSlaveData.TxDataLength)
    {
        // Transmit the next byte
        USART->TXDATA = (rsuint32)DrvSpiSlaveData.TxDataPtr[DrvSpiSlaveData.TxCount++];    
    }
    // No more data to send?
    if (DrvSpiSlaveData.TxCount == DrvSpiSlaveData.TxDataLength)
    {
      // Disable TX Buffer Low interrupt if no more bytes in queue
      USART_IntDisable(USART, USART_IEN_TXBL);
  
      // Notify application that all data has been successfully sent  
      DrvSpiSlaveTxDoneMailType *pM = (DrvSpiSlaveTxDoneMailType*)RosMailAllocate(COLA_TASK, COLA_TASK, sizeof(DrvSpiSlaveTxDoneMailType));
      pM->Primitive = SPI_SLAVE_TX_DONE;
      pM->Status = RSS_SUCCESS;
      pM->TxDataPtr = DrvSpiSlaveData.TxDataPtr;
      pM->TxDataLength = DrvSpiSlaveData.TxDataLength;
      RosMailDeliver((RosMailType*)pM);
      DrvSpiSlaveData.TxDataPtr = NULL;
      DrvSpiSlaveData.TxDataLength = 0;
      DrvSpiSlaveData.TxCount = 0;
    }      
  }  
}

// SPI Initialization thread
PT_THREAD(PtDrvSpiSlaveInit(struct pt *Pt, const RosMailType* Mail, rsuint32 BaudRate))
{
  PT_BEGIN(Pt);
  
  DrvSpiSlaveInit(BaudRate);
  DrvSpiSlaveData.TxDataLength = 0;
  
  PT_END(Pt);
}

// SPI Close thread
PT_THREAD(PtDrvSpiSlaveClose(struct pt *Pt, const RosMailType* Mail))
{
  PT_BEGIN(Pt);

  // Disable driver
  DrvSpiSlaveData.Enabled = FALSE;

  // Restore the original ISR handler
  SetIsr(USART_RX_INTERRUPT , NULL);  
  SetIsr(USART_TX_INTERRUPT , NULL);  
 
  PT_END(Pt);
}

// Initialization function
void DrvSpiSlaveInit(rsuint32 BaudRate)
{  
  USART_InitSync_TypeDef spiSlaveInit = USART_INITSYNC_DEFAULT;

  // Request the system clock manager to keep the system clk (CPU clk)
  // active. This prevents usage of EM2 and EM3 low power modes where
  // the HF clock source is switched off!
  ScCpuRequest(USART_SCU_DEVICE, SCS_ACTIVE);

  /* Enable TX and RX pin */
   
  GPIO_PinModeSet(gpioPortC, 2, gpioModeInput,    1); //MOSI
  GPIO_PinModeSet(gpioPortC, 3, gpioModePushPull, 1); //MISO
  GPIO_PinModeSet(gpioPortC, 4, gpioModeInput,0); //SPI_CLK as input
  GPIO_PinModeSet(gpioPortC, 5, gpioModeInput,1); //SPI_CS as input
      
  // Precautions pertaining to electrically connected ports on the RTX41xx
  GPIO_PinModeSet(gpioPortD, 4, gpioModeDisabled, 0); //Electrically connected to USART2_SPI_CLK, so disable
  GPIO_PinModeSet(gpioPortB, 11, gpioModeDisabled,0); //Electrically connected to USART2_SPI_CS, so disable
    
  // Enable clock for SPI module
  CMU_ClockEnable(cmuClock_USART2, true);
  
  // Set our own ISR handlers
  SetIsr(USART_RX_INTERRUPT , SpiSlaveRxIntHandler);  
  SetIsr(USART_TX_INTERRUPT , SpiSlaveTxIntHandler);  
  
  // Initialize USART 
  spiSlaveInit.baudrate = BaudRate;
  USART_InitSync(USART, &spiSlaveInit);  
  
  // Activate AUTO CHIP SELECT
  USART->CMD |= USART_CMD_MASTERDIS;
  
  // Enable Chip Select Invert (source:3Party/EnergyMicro/EFM32G/efm32g_usart.h)
  USART->CTRL |= USART_CTRL_CSINV;
  
  // Route USART RX and TX pins 
  USART->ROUTE = USART_ROUTE_RXPEN | USART_ROUTE_TXPEN | USART_ROUTE_CLKPEN | USART_ROUTE_CSPEN | USART_ROUTE_LOCATION ;
  
  // Enable RX interrupt
  USART_IntClear(USART, USART_IF_RXDATAV);
  NVIC_EnableIRQ(USART_RX_IRQN);
  USART_IntEnable(USART, USART_IF_RXDATAV);

  DrvSpiSlaveData.Enabled = TRUE;
  trigger_buffer_rx = 0;

}

// RX buffer read function
rsuint16 DrvSpiSlaveRx(rsuint8 *RxBufferPtr, rsuint16 RxBufferLength)
{
  rsuint16 n = RSMIN(DrvSpiSlaveRxGetSize(), RxBufferLength);
  
  // Clear RxPending flag to make sure we get notified when more data is received.
  DrvSpiSlaveData.RxPending = FALSE;

  // Any data to read?
  if (n)
  {
    // We have to handle wrapping in the RX buffer if ReadIdx > WriteIdx 
    if (DrvSpiSlaveData.RxReadIdx > DrvSpiSlaveData.RxWriteIdx)
    {
      
      rsuint16 n1 = RSMIN(n, SPI_SLAVE_RX_BUFFER_SIZE - DrvSpiSlaveData.RxReadIdx);

      // Read from ReadIdx to end of buffer 
      memcpy(RxBufferPtr, &SpiSlaveRxbuffer[DrvSpiSlaveData.RxReadIdx], n1);
      DrvSpiSlaveData.RxReadIdx += n1;
      if (DrvSpiSlaveData.RxReadIdx == SPI_SLAVE_RX_BUFFER_SIZE)
      {
        DrvSpiSlaveData.RxReadIdx = 0;
      }
      // Read from start of buffer to WriteId
      if (n1 < n)
      {
        n1 = n-n1;
        memcpy(RxBufferPtr, SpiSlaveRxbuffer, n1);
        DrvSpiSlaveData.RxReadIdx += n1;
      }
    }
    else //if (DrvSpiSlaveData.RxReadIdx < DrvSpiSlaveData.RxWriteIdx)
    {
      // Just copy data if ReadIdx < WriteIdx
      memcpy(RxBufferPtr, &SpiSlaveRxbuffer[DrvSpiSlaveData.RxReadIdx], n);
      DrvSpiSlaveData.RxReadIdx += n;
      if (DrvSpiSlaveData.RxReadIdx == SPI_SLAVE_RX_BUFFER_SIZE)
      {
        DrvSpiSlaveData.RxReadIdx = 0;
      }
    }
  }  
  return n;
}

// Function to Flush out the RX buffer 
void DrvSpiSlaveRxFlush(void)
{
  // We flush the RX buffer by resetting both WriteIdx and ReadIdx
  // Disable interrupt first to make sure that we do not have an 
  // RX interrupt while we are messing around with WriteIdx and ReadIdx. 
  RsIntStatusType saved = RsDisableInterrupts(); 
  DrvSpiSlaveData.RxReadIdx = 0;
  DrvSpiSlaveData.RxWriteIdx = 0;
  DrvSpiSlaveData.RxPending = FALSE;

  // Enable interrupts again
  RsEnableInterrupts(saved); 
}

// Start transmission
RsStatusType DrvSpiSlaveTxStart(rsuint8 *TxDataPtr, rsuint16 TxDataLength)
{
  if (TxDataPtr == NULL || TxDataLength == 0)
  {
    return RSS_BAD_ARGUMENTS;
  }

  if (DrvSpiSlaveData.TxDataLength == 0)
  {
    // Store pointer to data
    DrvSpiSlaveData.TxDataPtr = TxDataPtr;
    DrvSpiSlaveData.TxDataLength = TxDataLength;
    DrvSpiSlaveData.TxCount = 0;
    
    // TX buffer low interrupt on half empty TX buffer
    USART->CTRL |= USART_CTRL_TXBIL; 
    
    // Enable TX buffer low interrupt
    NVIC_EnableIRQ(USART_TX_IRQN);
    USART_IntEnable(USART, USART_IEN_TXBL);
    USART_IntDisable(USART, USART_IEN_TXC);
      
    
    return RSS_SUCCESS;
  }
  return RSS_BUSY;
}

void DrvSpiSlaveSetTrigger(rsuint16 l)
{
    trigger_buffer_rx = l;
}

rsuint16 DrvSpiSlaveGetBufferSize()
{
    return (SPI_SLAVE_RX_BUFFER_SIZE-1);
}
